import copy
from saml2 import samlp, SamlBase
from saml2 import NAMEID_FORMAT_EMAILADDRESS
from saml2 import BINDING_HTTP_REDIRECT
from saml2 import BINDING_HTTP_POST
from saml2.s_utils import rndstr

from saml2.saml import SCM_BEARER, Condition, XSI_TYPE, Audience
from saml2.saml import NAMEID_FORMAT_PERSISTENT
from saml2.saml import SCM_SENDER_VOUCHES
from saml2.saml import ConditionAbstractType_
from saml2.samlp import STATUS_AUTHN_FAILED
from saml2.time_util import in_a_while, a_while_ago
from sp_test import check
from sp_test.check import VerifyAuthnRequest, VerifyDigestAlgorithm
from sp_test.check import VerifySignatureAlgorithm, VerifyIfRequestIsSigned
from sp_test.check import SetResponseAndAssertionSignaturesFalse
from saml2test.check import CheckSpHttpResponseOK, CheckSpHttpResponse500
from saml2test import ip_addresses

__author__ = 'rolandh'

USER = {
    "adam": {
        "given_name": "Adam",
        "sn": "Andersson"
    },
    "eva": {
        "given_name": "Eva",
        "sn": "Svensson"
    }
}


# Extension class - extra condition
class TimeRestriction(ConditionAbstractType_):
    """ """

    c_tag = 'TimeRestriction'
    c_namespace = "urn:mace:umu.se:sso"
    c_children = ConditionAbstractType_.c_children.copy()
    c_attributes = ConditionAbstractType_.c_attributes.copy()
    c_child_order = ConditionAbstractType_.c_child_order[:]
    c_cardinality = ConditionAbstractType_.c_cardinality.copy()
    c_attributes['StartTime'] = ('start_time', 'time', False)
    c_attributes['EndTime'] = ('end_time', 'time', False)

    def __init__(self,
                 start_time=None,
                 end_time=None,
                 text=None,
                 extension_elements=None,
                 extension_attributes=None):
        ConditionAbstractType_.__init__(
            self, text=text, extension_elements=extension_elements,
            extension_attributes=extension_attributes)
        self.start_time = start_time
        self.end_time = end_time


# =============================================================================


class Response(object):
    _args = {}
    _class = samlp.Response
    _sign = False
    tests = {"pre": [], "post": []}

    def __init__(self, conv):
        self.args = self._args.copy()
        self.conv = conv

    def setup(self):
        pass

    def pre_processing(self, message, *kwargs):
        return message

    def post_processing(self, message, *kwargs):
        return message


class Request(object):
    response = ""
    _class = None
    tests = {"pre": [],
             "mid": [VerifyAuthnRequest],
             "post": []}

    def __init__(self):
        pass

    def __call__(self, conv, response):
        pass


class Operation(object):
    pass


class AuthnResponse(Response):
    _response_args = {
        "identity": USER["adam"],
        "userid": "adam",
    }
    _binding = BINDING_HTTP_POST


class AuthnResponse_redirect(AuthnResponse):
    _binding = BINDING_HTTP_REDIRECT


class ErrorResponse(Response):
    _response_args = {
        "info": (STATUS_AUTHN_FAILED, "Unknown user")
    }
    _binding = BINDING_HTTP_POST


class LogoutResponse(Response):
    _class = samlp.LogoutRequest
    pass


class Login(Operation):
    _interaction = ["wayf"]


class AuthnRequest(Request):
    _class = samlp.AuthnRequest


class AuthnResponse_NameIDformat_persistent(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        name_id = message.assertion.subject.name_id
        name_id.name_format = NAMEID_FORMAT_PERSISTENT
        return message


class AuthnResponse_NameIDformat_email(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        name_id = message.assertion.subject.name_id
        name_id.name_format = NAMEID_FORMAT_EMAILADDRESS
        name_id.text = "adam@example.com"
        return message


class AuthnResponse_NameIDformat_foo(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        name_id = message.assertion.subject.name_id
        name_id.name_format = "foo"
        name_id.text = "fruit basket"
        return message


class AuthnResponse_without_SubjectConfirmationData_1(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        _confirmation.subject_confirmation_data = None
        _confirmation.method = SCM_SENDER_VOUCHES
        return message


class AuthnResponse_without_SubjectConfirmationData_2(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        _confirmation.subject_confirmation_data = None
        _confirmation.method = SCM_BEARER
        return message


class AuthnResponse_rnd_Response_inresponseto(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.in_response_to = "invalid_rand_" + rndstr(6)
        return message


class AuthnResponse_rnd_Response_assertion_inresponseto(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.assertion.in_response_to = "invalid_rand_" + rndstr(6)
        return message


class AuthnResponse_Response_no_inresponse(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.in_response_to = None
        return message


class AuthnResponse_SubjectConfirmationData_no_inresponse(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        _confirmation[0].subject_confirmation_data.in_response_to = None
        return message


class AuthnResponse_wrong_Recipient(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        _confirmation[0].subject_confirmation_data.recipient = rndstr(16)
        return message


class AuthnResponse_missing_Recipient(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        _confirmation[0].subject_confirmation_data.recipient = None
        return message


class AuthnResponse_missing_Recipient(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        _confirmation[0].subject_confirmation_data.recipient = None
        return message


class AuthnResponse_broken_destination(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.destination = "NotAUrl"
        return message


class AuthnResponse_correct_recipient_address(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        if "localhost" in self.conv.entity_id:
            addr = "127.0.0.1"
        else:
            addr = ip_addresses()[0]
        _confirmation[0].subject_confirmation_data.address = addr
        return message


class AuthnResponse_incorrect_recipient_address(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        _confirmation[0].subject_confirmation_data.address = "10.0.0.1"
        return message


class AuthnResponse_2_recipients_me_last(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        sc = copy.copy(_confirmation[0])
        if "localhost" in self.conv.entity_id:
            addr = "127.0.0.1"
        else:
            addr = ip_addresses()[0]
        sc.subject_confirmation_data.address = addr
        _confirmation.insert(0, sc)
        return message


class AuthnResponse_2_recipients_me_first(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation
        sc = copy.copy(_confirmation[0])
        if "localhost" in self.conv.entity_id:
            addr = "127.0.0.1"
        else:
            addr = ip_addresses()[0]
        sc.subject_confirmation_data.address = addr
        _confirmation.append(sc)
        return message


class AuthnResponse_unknown_condition(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        conditions = message.assertion.conditions
        conditions.condition = [Condition(
            extension_elements=[TimeRestriction(start_time="08:00:00",
                                                end_time="17:00:00")],
            extension_attributes={XSI_TYPE: "foo:bas"})]
        return message


class AuthnResponse_future_NotBefore(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        conditions = message.assertion.conditions
        # Valid starting five hours from now
        conditions.not_before = in_a_while(hours=5)
        return message


class AuthnResponse_past_NotOnOrAfter(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        conditions = message.assertion.conditions
        # Valid up until five hours ago
        conditions.not_on_or_after = a_while_ago(hours=5)
        return message


class AuthnResponse_past_SubjectConfirmationData_NotOnOrAfter(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation[0]
        _confirmation.subject_confirmation_data.not_on_or_after = a_while_ago(
            hours=5)
        return message


class AuthnResponse_future_SubjectConfirmationData_NotBefore(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _confirmation = message.assertion.subject.subject_confirmation[0]
        _confirmation.subject_confirmation_data.not_before = in_a_while(
            hours=5)
        return message


class AuthnResponse_past_AuthnStatement_SessionNotOnOrAfter(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        _statement = message.assertion.authn_statement[0]
        _statement.session_not_on_or_after = a_while_ago(hours=5)
        return message


class AuthnResponse_missing_AuthnStatement(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.assertion.authn_statement = []
        return message


class AuthnResponse_future_24h_IssueInstant(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.assertion.issue_instant = in_a_while(hours=24)
        return message


class AuthnResponse_past_24h_IssueInstant(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.assertion.issue_instant = a_while_ago(hours=24)
        return message


class AuthnResponse_datetime_millisecond(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        message.assertion.issue_instant = in_a_while(milliseconds=123)
        return message


class AuthnResponse_AudienceRestriction_no_audience(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        conditions = message.assertion.conditions
        conditions.audience_restriction[0].audience = None
        return message


class AuthnResponse_AudienceRestriction_wrong_audience(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        conditions = message.assertion.conditions
        conditions.audience_restriction[0].audience = [
            Audience("http://saml.example.com")]
        return message


class AuthnResponse_AudienceRestriction_prepended_audience(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        conditions = message.assertion.conditions
        extra = Audience("http://saml.example.com")
        conditions.audience_restriction[0].audience.insert(0, extra)
        return message


class AuthnResponse_AudienceRestriction_appended_audience(AuthnResponse):
    def pre_processing(self, message, **kwargs):
        conditions = message.assertion.conditions
        extra = Audience("http://saml.example.com")
        conditions.audience_restriction[0].audience.append(extra)
        return message


PHASES = {
    "login_redirect": (Login, AuthnRequest, AuthnResponse_redirect),
}

# Each operation defines 4 flows and 3 sets of tests, in chronological order:
# test "pre": executes before anything is sent to the SP
# flow 0: Start conversation flow
# flow 1: SAML request flow
# test "mid": executes after receiving the SAML request
# flow 2: SAML response flow
# flow 3: check SP response after authentication
# test "post": executes after finals response has been received from SP

OPERATIONS = {
    'sp-00': {
        "name": 'Basic Login test expect HTTP 200 result',
        "descr": 'WebSSO verify authentication request, verify '
                 'HTTP-Response after sending the SAML response',
        "sequence": [(Login, AuthnRequest, AuthnResponse, CheckSpHttpResponseOK)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'sp-01': {
        "name": 'Login OK & echo page verification test',
        "descr": 'Same as SP-00, then check if result page is displayed',
        "sequence": [(Login, AuthnRequest, AuthnResponse, check.VerifyEchopageContents)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'sp-02': {
        "name": 'Require AuthnRequest to be signed',
        "descr": 'Same as SP-00, and check if a request signature can be found',
        "sequence": [(Login, AuthnRequest, AuthnResponse, None)],
        "tests": {"pre": [], "mid": [VerifyIfRequestIsSigned], "post": []}
    },
    'sp-03': {
        "name": 'Reject unsigned reponse/assertion',
        "descr": 'Check if SP flags missing signature with HTTP 500',
        "sequence": [(Login, AuthnRequest, AuthnResponse, CheckSpHttpResponse500)],
        "tests": {"pre": [SetResponseAndAssertionSignaturesFalse], "mid": [], "post": []}
    },
    'sp-04': {  # test-case specific code in sp_test/__init__
        "name": 'Reject siganture with invalid IDP key',
        "descr": 'IDP-key for otherwise valid signature not in metadata - expect HTTP 500 result',
        "sequence": [(Login, AuthnRequest, AuthnResponse, CheckSpHttpResponse500)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'sp-05': {
        "name": 'Verify digest algorithm',
        "descr": 'Trigger WebSSO AuthnRequest and verify that the used '
                 'digest algorithm was one from the approved set.',
        "sequence": [(Login, AuthnRequest, AuthnResponse, None)],
        "tests": {"pre": [], "mid": [VerifyDigestAlgorithm], "post": []}
    },
    'sp-06': {
        "name": 'Verify signature algorithm',
        "descr": 'Trigger WebSSO AuthnRequest and verify that the used '
                 'signature algorithm was one from the approved set.',
        "sequence": [(Login, AuthnRequest, AuthnResponse, None)],
        "tests": {"pre": [], "mid": [VerifySignatureAlgorithm], "post": []}
    },
    'sp-08': {
        "name": "SP should accept a Response without a "
                 "SubjectConfirmationData element. If confirmation method"
                 "is SCM_SENDER_VOUCHES",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_without_SubjectConfirmationData_2,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL02': {
        "name": 'Verify various aspects of the generated AuthnRequest message',
        "descr": 'Basic Login test',
        "sequence": [],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL03': {
        "name": "SP should not accept a Response as valid, when the StatusCode"
                " is not success",
        "sequence": [(Login, AuthnRequest, ErrorResponse, check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL04': {
        "name": "SP should accept a NameID with Format: persistent",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_NameIDformat_persistent, None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL05': {
        "name": "SP should accept a NameID with Format: e-mail",
        "sequence": [(Login, AuthnRequest, AuthnResponse_NameIDformat_email,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL06': {
        "name": "Do SP work with unknown NameID Format, such as : foo",
        "sequence": [(Login, AuthnRequest, AuthnResponse_NameIDformat_foo,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL07': {
        "name": "SP should accept a Response without a "
                 "SubjectConfirmationData element. If confirmation method "
                 "is SCM_SENDER_VOUCHES",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_without_SubjectConfirmationData_1, None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL09': {
        "name": "SP should not accept a response InResponseTo "
                "which is chosen randomly",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_rnd_Response_inresponseto,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL10': {
        "name": "SP should not accept an assertion InResponseTo "
                 "which is chosen randomly",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_rnd_Response_assertion_inresponseto,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL11': {
        "name": "Does the SP allow the InResponseTo attribute to be missing"
                "from the Response element?",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_Response_no_inresponse,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL12': {
        "name": "Does the SP allow the InResponseTo attribute to be missing"
                "from the SubjectConfirmationData element?"
                "(Test is questionable - review)",  # TODO
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_SubjectConfirmationData_no_inresponse,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL13': {
        "name": "SP should not accept a broken DestinationURL attribute",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_broken_destination,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    # New untested
    'FL14a': {
        "name": "SP should not accept wrong Recipient attribute",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_broken_destination,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL14b': {
        "name": "SP should not accept missing Recipient attribute",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_missing_Recipient,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL20': {
        "name": "Accept a Response with a SubjectConfirmationData elements "
                "with a correct @Address attribute",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_correct_recipient_address,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL21': {
        "name": "Accept a Response with a SubjectConfirmationData elements "
                "with a incorrect @Address attribute",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_incorrect_recipient_address,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL22': {
        "name": "Accept a Response with two SubjectConfirmationData elements"
                "representing two recipients (test 1 of 2, correct one last)",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_2_recipients_me_last,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL23': {
        "name": "Accept a Response with two SubjectConfirmationData elements"
                "representing two recipients (test 1 of 2, correct one last)",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_2_recipients_me_first,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL26': {
        "name": "Reject an assertion containing an unknown Condition.",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_unknown_condition,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL27': {
        "name": "Reject a Response with a Condition with a NotBefore in the "
                "future.",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_future_NotBefore,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL28': {
        "name": "Reject a Response with a Condition with a NotOnOrAfter in "
                "the past.",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_future_NotBefore,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL29': {
        "name": "Reject a Response with a SubjectConfirmationData@NotOnOrAfter "
                "in the past",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_past_SubjectConfirmationData_NotOnOrAfter,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL24': {
        "name": "Reject a Response with a SubjectConfirmationData@NotBefore "
                "in the future",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_future_SubjectConfirmationData_NotBefore,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL30': {
        "name": "Reject a Response with an AuthnStatement where "
                "SessionNotOnOrAfter is set in the past.",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_past_AuthnStatement_SessionNotOnOrAfter,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL31': {
        "name": "Reject a Response with an AuthnStatement missing",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_missing_AuthnStatement,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL32': {
        "name": "Reject an IssueInstant far (24 hours) into the future",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_future_24h_IssueInstant,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL33': {
        "name": "Reject an IssueInstant far (24 hours) into the past",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_past_24h_IssueInstant,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL34': {
        "name": "Accept xs:datetime with millisecond precision "
                "http://www.w3.org/TR/xmlschema-2/#dateTime",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_datetime_millisecond,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL36': {
        "name": "Reject a Response with a Condition with a empty set of "
                "Audience.",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_AudienceRestriction_no_audience,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL37': {
        "name": "Reject a Response with a Condition with a wrong Audience.",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_AudienceRestriction_wrong_audience,
                      check.ErrorResponse)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL38': {
        "name": "Accept a Response with a Condition with an additional "
                "Audience prepended",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_AudienceRestriction_prepended_audience,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
    'FL39': {
        "name": "Accept a Response with a Condition with an additional "
                "Audience appended",
        "sequence": [(Login, AuthnRequest,
                      AuthnResponse_AudienceRestriction_appended_audience,
                      None)],
        "tests": {"pre": [], "mid": [], "post": []}
    },
}

#
# SP should not accept a broken Recipient attribute in assertion
# SubjectConfirmationData/@Recipient
# SP should not accept a broken DestinationURL attribute in response
# SP should accept a Response with two SubjectConfirmationData elements
# representing two recipients (test 1 of 2, correct one last)
# SP should accept a Response with two SubjectConfirmationData elements
# representing two recipients (test 1 of 2, correct one first)
# SP should accept a Response with two SubjectConfirmation elements
# representing two recipients (test 1 of 2, correct one last)
# SP should accept a Response with two SubjectConfirmation elements
# representing two recipients (test 1 of 2, correct one first)
# SP should accept a Response with a SubjectConfirmationData elements with a
#  correct @Address attribute
# SP should nnot accept a Response with a SubjectConfirmationData elements
# with a incorrect @Address attribute
# SP should accept a Response with multiple SubjectConfirmation elements
# with /SubjectConfirmationData/@Address-es, where one is correct (test 1 of
# 2, correct o last)
# SP should accept a Response with multiple SubjectConfirmationData elements
#  with /SubjectConfirmationData/@Address-es, where one is correct (test 1 of
#  2, corr one last)
# SP should accept a Response with multiple SubjectConfirmationData elements
#  with /SubjectConfirmationData/@Address-es, where one is correct (test 1 of
#  2, corr one first)
# SP Should not accept an assertion containing an uknown Condition
# SP should not accept a Response with a Condition with a NotBefore in the
# future.
# SP should not accept a Response with a Condition with a NotOnOrAfter in
# the past.
# SP should not accept a Response with a
# SubjectConfirmationData@NotOnOrAfter in the past
# SP should not accept a Response with a AuthnStatement where
# SessionNotOnOrAfter is set in the past
# SP should not accept a Response with a AuthnStatement missing
# SP should not accept an IssueInstant far (24 hours) into the future
# SP should not accept an IssueInstant far (24 hours) into the past
# SP should accept xs:datetime with millisecond precision http://www.w3
# .org/TR/xmlschema-2/#dateTime
# SP should accept xs:datetime with microsecond precision http://www.w3
# .org/TR/xmlschema-2/#dateTime
# SP should not accept a Response with a Condition with a empty set of
# Audience.
# SP should not accept a Response with a Condition with a wrong Audience.
# SP should accept a Response with a Condition with an addition Audience
# prepended.
# SP should accept a Response with a Condition with an addition Audience
# appended.
# SP should not accept multiple AudienceRestrictions where the intersection
# is zero. (test 1 of 2)
# SP should not accept multiple AudienceRestrictions where the intersection
# is zero. (test 2 of 2)
# SP should accept multiple AudienceRestrictions where the intersection
# includes the correct audience.
# SP should accept that only the Assertion is signed instead of the Response.
# SP should accept that both the Response and the Assertion is signed.
# Do SP work when RelayState information is lost?
# Do SP accept an unknown Extensions element in the Response?
# SP MUST not accept response when the saml-namespace is invalid
# SP MUST NOT re-use the same ID in subsequent requests.
# SP MUST NOT accept a replayed Response. An identical Response/Assertion
# used a second time. [Profiles]: 4.1.4.5 POST-Specific Processing Rules (
# test 1 of 2: s inresponseto)
# SP MUST NOT accept a replayed Response. An identical Response/Assertion
# used a second time. [Profiles]: 4.1.4.5 POST-Specific Processing Rules (
# test 2 of 2: unsolicited response)
# SP SHOULD find attributes in a second AttributeStatement, not only in the
# first.
# SP SHOULD NOT accept an signed assertion embedded in an AttributeValue
# inside an unsigned assertion.
# SP SHOULD NOT accept an signed assertion embedded in an AttributeValue
# inside an unsigned assertion. (Signature moved out...)
# SP SHOULD NOT accept an signed assertion, where the signature is referring
#  to another assertion.
# SP SHOULD find attributes in a second Assertion/AttributeStatement,
# not only in one of them (test 1 of 2 - attributes in first).
# SP SHOULD find attributes in a second Assertion/AttributeStatement,
# not only in one of them (test 2 of 2 - attributes in last).
# SP SHOULD NOT accept attributes in unsigned 2nd assertion. (test 1 of 2)
# SP SHOULD NOT accept attributes in unsigned 2nd assertion. (test 2 of 2)
# SP SHOULD NOT accept authnstatement in unsigned 2nd assertion. (test 1 of 2)
# SP SHOULD NOT accept authnstatement in unsigned 2nd assertion. (test 2 of 2)
# Basic SP-initated Logout Test
# Basic IdP-initated Logout Test
# SP MUST NOT accept LogoutRequest when NameID content is wrong
# SP MUST NOT accept LogoutRequest when NameID@Format is wrong
# SP MUST NOT accept LogoutRequest when NameID@SPNameQualifier is wrong
# SP MUST NOT logout user when invalid SessionIndex is sent
# SP MUST NOT accept LogoutRequest when Issuer is wrong
# SP MUST NOT accept LogoutRequest when Destination is wrong
# SP MUST NOT accept unsigned LogoutRequest
# SP MUST accept LogoutRequest with sessionindex in a separate session,
# ot relying on the session-cookie.
# SP MUST accept an LogoutRequest with no sessionindex (sent in separate
# session, no session-cookies)
# SP MUST accept an LogoutRequest with two sesionindexes (first valid) (sent
#  in separate session, no session-cookies)
# SP MUST accept an LogoutRequest with two sesionindexes (second valid) (
# sent in separate session, no session-cookies)
# Session fixtation check


from saml2.assertion import Policy

HOST = '127.0.0.1'
PORT = 8087
HTTPS = False

# Which groups of entity categories to use
POLICY = Policy(
    {
        "default": {"entity_categories": ["swamid", "edugain"]}
    }
)

# HTTPS cert information
SERVER_CERT = "pki/mycert.pem"
SERVER_KEY = "pki/mykey.pem"
CERT_CHAIN = ""



#!/usr/bin/env python
import logging
import random

import time
import base64
import six
import sys
import hmac
import string

# from python 2.5
import imp
import traceback

if sys.version_info >= (2, 5):
    import hashlib
else:  # before python 2.5
    import sha

from saml2 import saml
from saml2 import samlp
from saml2 import VERSION
from saml2.time_util import instant

try:
    from hashlib import md5
except ImportError:
    from md5 import md5
import zlib

logger = logging.getLogger(__name__)


class SamlException(Exception):
    pass


class RequestVersionTooLow(SamlException):
    pass


class RequestVersionTooHigh(SamlException):
    pass


class UnknownPrincipal(SamlException):
    pass


class UnknownSystemEntity(SamlException):
    pass


class Unsupported(SamlException):
    pass


class UnsupportedBinding(Unsupported):
    pass


class VersionMismatch(Exception):
    pass


class Unknown(Exception):
    pass


class OtherError(Exception):
    pass


class MissingValue(Exception):
    pass


class PolicyError(Exception):
    pass


class BadRequest(Exception):
    pass


class UnravelError(Exception):
    pass


EXCEPTION2STATUS = {
    VersionMismatch: samlp.STATUS_VERSION_MISMATCH,
    UnknownPrincipal: samlp.STATUS_UNKNOWN_PRINCIPAL,
    UnsupportedBinding: samlp.STATUS_UNSUPPORTED_BINDING,
    RequestVersionTooLow: samlp.STATUS_REQUEST_VERSION_TOO_LOW,
    RequestVersionTooHigh: samlp.STATUS_REQUEST_VERSION_TOO_HIGH,
    OtherError: samlp.STATUS_UNKNOWN_PRINCIPAL,
    MissingValue: samlp.STATUS_REQUEST_UNSUPPORTED,
    # Undefined
    Exception: samlp.STATUS_AUTHN_FAILED,
}

GENERIC_DOMAINS = ["aero", "asia", "biz", "cat", "com", "coop", "edu",
                   "gov", "info", "int", "jobs", "mil", "mobi", "museum",
                   "name", "net", "org", "pro", "tel", "travel"]


def valid_email(emailaddress, domains=GENERIC_DOMAINS):
    """Checks for a syntactically valid email address."""

    # Email address must be at least 6 characters in total.
    # Assuming noone may have addresses of the type a@com
    if len(emailaddress) < 6:
        return False  # Address too short.

    # Split up email address into parts.
    try:
        localpart, domainname = emailaddress.rsplit('@', 1)
        host, toplevel = domainname.rsplit('.', 1)
    except ValueError:
        return False  # Address does not have enough parts.

    # Check for Country code or Generic Domain.
    if len(toplevel) != 2 and toplevel not in domains:
        return False  # Not a domain name.

    for i in '-_.%+.':
        localpart = localpart.replace(i, "")
    for i in '-_.':
        host = host.replace(i, "")

    if localpart.isalnum() and host.isalnum():
        return True  # Email address is fine.
    else:
        return False  # Email address has funny characters.


def decode_base64_and_inflate(string):
    """ base64 decodes and then inflates according to RFC1951

    :param string: a deflated and encoded string
    :return: the string after decoding and inflating
    """

    return zlib.decompress(base64.b64decode(string), -15)


def deflate_and_base64_encode(string_val):
    """
    Deflates and the base64 encodes a string

    :param string_val: The string to deflate and encode
    :return: The deflated and encoded string
    """
    if not isinstance(string_val, six.binary_type):
        string_val = string_val.encode('utf-8')
    return base64.b64encode(zlib.compress(string_val)[2:-4])


def rndstr(size=16, alphabet=""):
    """
    Returns a string of random ascii characters or digits

    :param size: The length of the string
    :return: string
    """
    rng = random.SystemRandom()
    if not alphabet:
        alphabet = string.ascii_letters[0:52] + string.digits
    return type(alphabet)().join(rng.choice(alphabet) for _ in range(size))

def rndbytes(size=16, alphabet=""):
    """
    Returns rndstr always as a binary type
    """
    x = rndstr(size, alphabet)
    if isinstance(x, six.string_types):
        return x.encode('utf-8')
    return x


def sid():
    """creates an unique SID for each session.
    160-bits long so it fulfills the SAML2 requirements which states
    128-160 bits

    :return: A random string prefix with 'id-' to make it
        compliant with the NCName specification
    """
    return "id-" + rndstr(17)


def parse_attribute_map(filenames):
    """
    Expects a file with each line being composed of the oid for the attribute
    exactly one space, a user friendly name of the attribute and then
    the type specification of the name.

    :param filenames: List of filenames on mapfiles.
    :return: A 2-tuple, one dictionary with the oid as keys and the friendly
        names as values, the other one the other way around.
    """
    forward = {}
    backward = {}
    for filename in filenames:
        for line in open(filename).readlines():
            (name, friendly_name, name_format) = line.strip().split()
            forward[(name, name_format)] = friendly_name
            backward[friendly_name] = (name, name_format)

    return forward, backward


def identity_attribute(form, attribute, forward_map=None):
    if form == "friendly":
        if attribute.friendly_name:
            return attribute.friendly_name
        elif forward_map:
            try:
                return forward_map[(attribute.name, attribute.name_format)]
            except KeyError:
                return attribute.name
    # default is name
    return attribute.name

#----------------------------------------------------------------------------


def error_status_factory(info):
    if isinstance(info, Exception):
        try:
            exc_val = EXCEPTION2STATUS[info.__class__]
        except KeyError:
            exc_val = samlp.STATUS_AUTHN_FAILED
        try:
            msg = info.args[0]
        except IndexError:
            msg = "%s" % info
        status = samlp.Status(
            status_message=samlp.StatusMessage(text=msg),
            status_code=samlp.StatusCode(
                value=samlp.STATUS_RESPONDER,
                status_code=samlp.StatusCode(
                    value=exc_val)))
    else:
        (errcode, text) = info
        status = samlp.Status(
            status_message=samlp.StatusMessage(text=text),
            status_code=samlp.StatusCode(
                value=samlp.STATUS_RESPONDER,
                status_code=samlp.StatusCode(value=errcode)))

    return status


def success_status_factory():
    return samlp.Status(status_code=samlp.StatusCode(
        value=samlp.STATUS_SUCCESS))


def status_message_factory(message, code, fro=samlp.STATUS_RESPONDER):
    return samlp.Status(
        status_message=samlp.StatusMessage(text=message),
        status_code=samlp.StatusCode(value=fro,
                                     status_code=samlp.StatusCode(value=code)))


def assertion_factory(**kwargs):
    assertion = saml.Assertion(version=VERSION, id=sid(),
                               issue_instant=instant())
    for key, val in kwargs.items():
        setattr(assertion, key, val)
    return assertion


def _attrval(val, typ=""):
    if isinstance(val, list) or isinstance(val, set):
        attrval = [saml.AttributeValue(text=v) for v in val]
    elif val is None:
        attrval = None
    else:
        attrval = [saml.AttributeValue(text=val)]

    if typ:
        for ava in attrval:
            ava.set_type(typ)

    return attrval

# --- attribute profiles -----

# xmlns:xs="http://www.w3.org/2001/XMLSchema"
# xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"


def do_ava(val, typ=""):
    if isinstance(val, six.string_types):
        ava = saml.AttributeValue()
        ava.set_text(val)
        attrval = [ava]
    elif isinstance(val, list):
        attrval = [do_ava(v)[0] for v in val]
    elif val or val is False:
        ava = saml.AttributeValue()
        ava.set_text(val)
        attrval = [ava]
    elif val is None:
        attrval = None
    else:
        raise OtherError("strange value type on: %s" % val)

    if typ:
        for ava in attrval:
            ava.set_type(typ)

    return attrval


def do_attribute(val, typ, key):
    attr = saml.Attribute()
    attrval = do_ava(val, typ)
    if attrval:
        attr.attribute_value = attrval

    if isinstance(key, six.string_types):
        attr.name = key
    elif isinstance(key, tuple):  # 3-tuple or 2-tuple
        try:
            (name, nformat, friendly) = key
        except ValueError:
            (name, nformat) = key
            friendly = ""
        if name:
            attr.name = name
        if format:
            attr.name_format = nformat
        if friendly:
            attr.friendly_name = friendly
    return attr


def do_attributes(identity):
    attrs = []
    if not identity:
        return attrs
    for key, spec in identity.items():
        try:
            val, typ = spec
        except ValueError:
            val = spec
            typ = ""
        except TypeError:
            val = ""
            typ = ""

        attr = do_attribute(val, typ, key)
        attrs.append(attr)
    return attrs


def do_attribute_statement(identity):
    """
    :param identity: A dictionary with fiendly names as keys
    :return:
    """
    return saml.AttributeStatement(attribute=do_attributes(identity))


def factory(klass, **kwargs):
    instance = klass()
    for key, val in kwargs.items():
        setattr(instance, key, val)
    return instance


def signature(secret, parts):
    """Generates a signature. All strings are assumed to be utf-8
    """
    if not isinstance(secret, six.binary_type):
        secret = secret.encode('utf-8')
    newparts = []
    for part in parts:
        if not isinstance(part, six.binary_type):
            part = part.encode('utf-8')
        newparts.append(part)
    parts = newparts
    if sys.version_info >= (2, 5):
        csum = hmac.new(secret, digestmod=hashlib.sha1)
    else:
        csum = hmac.new(secret, digestmod=sha)

    for part in parts:
        csum.update(part)

    return csum.hexdigest()


def verify_signature(secret, parts):
    """ Checks that the signature is correct """
    if signature(secret, parts[:-1]) == parts[-1]:
        return True
    else:
        return False


FTICKS_FORMAT = "F-TICKS/SWAMID/2.0%s#"


def fticks_log(sp, logf, idp_entity_id, user_id, secret, assertion):
    """
    'F-TICKS/' federationIdentifier '/' version *('#' attribute '=' value) '#'
    Allowed attributes:
        TS	the login time stamp
        RP	the relying party entityID
        AP	the asserting party entityID (typcially the IdP)
        PN	a sha256-hash of the local principal name and a unique key
        AM	the authentication method URN

    :param sp: Client instance
    :param logf: The log function to use
    :param idp_entity_id: IdP entity ID
    :param user_id: The user identifier
    :param secret: A salt to make the hash more secure
    :param assertion: A SAML Assertion instance gotten from the IdP
    """
    csum = hmac.new(secret, digestmod=hashlib.sha1)
    csum.update(user_id)
    ac = assertion.AuthnStatement[0].AuthnContext[0]

    info = {
        "TS": time.time(),
        "RP": sp.entity_id,
        "AP": idp_entity_id,
        "PN": csum.hexdigest(),
        "AM": ac.AuthnContextClassRef.text
    }
    logf.info(FTICKS_FORMAT % "#".join(["%s=%s" % (a, v) for a, v in info]))


def dynamic_importer(name, class_name=None):
    """
    Dynamically imports modules / classes
    """
    try:
        fp, pathname, description = imp.find_module(name)
    except ImportError:
        print("unable to locate module: " + name)
        return None, None

    try:
        package = imp.load_module(name, fp, pathname, description)
    except Exception:
        raise

    if class_name:
        try:
            _class = imp.load_module("%s.%s" % (name, class_name), fp,
                                     pathname, description)
        except Exception:
            raise

        return package, _class
    else:
        return package, None


def exception_trace(exc):
    message = traceback.format_exception(*sys.exc_info())

    try:
        _exc = "Exception: %s" % exc
    except UnicodeEncodeError:
        _exc = "Exception: %s" % exc.message.encode("utf-8", "replace")

    return {"message": _exc, "content": "".join(message)}


def rec_factory(cls, **kwargs):
    _inst = cls()
    for key, val in kwargs.items():
        if key in ["text", "lang"]:
            setattr(_inst, key, val)
        elif key in _inst.c_attributes:
            try:
                val = str(val)
            except Exception:
                continue
            else:
                setattr(_inst, _inst.c_attributes[key][0], val)
        elif key in _inst.c_child_order:
            for tag, _cls in _inst.c_children.values():
                if tag == key:
                    if isinstance(_cls, list):
                        _cls = _cls[0]
                        claim = []
                        if isinstance(val, list):
                            for v in val:
                                claim.append(rec_factory(_cls, **v))
                        else:
                            claim.append(rec_factory(_cls, **val))
                    else:
                        claim = rec_factory(_cls, **val)
                    setattr(_inst, key, claim)
                    break

    return _inst



__author__ = 'rolandh'


NAME = ["givenName", "displayName", "sn"]
STATIC_ORG_INFO = ["c", "o", "co", "norEduOrgAcronym", "schacHomeOrganization"]
OTHER = ["eduPersonPrincipalName", "eduPersonScopedAffiliation", "mail"]

# These give you access to information
RESEARCH_AND_EDUCATION = "http://www.swamid.se/category/research-and-education"
SFS_1993_1153 = "http://www.swamid.se/category/sfs-1993-1153"

# presently these don't by themself
EU = "http://www.swamid.se/category/eu-adequate-protection"
NREN = "http://www.swamid.se/category/nren-service"
HEI = "http://www.swamid.se/category/hei-service"

RELEASE = {
    "": ["eduPersonTargetedID"],
    SFS_1993_1153: ["norEduPersonNIN"],
    (RESEARCH_AND_EDUCATION, EU): NAME + STATIC_ORG_INFO + OTHER,
    (RESEARCH_AND_EDUCATION, NREN): NAME + STATIC_ORG_INFO + OTHER,
    (RESEARCH_AND_EDUCATION, HEI): NAME + STATIC_ORG_INFO + OTHER,
}


from hashlib import sha1
import logging

from pymongo import MongoClient
from pymongo.mongo_replica_set_client import MongoReplicaSetClient
import pymongo.uri_parser
import pymongo.errors
from saml2.eptid import Eptid
from saml2.mdstore import InMemoryMetaData
from saml2.s_utils import PolicyError

from saml2.ident import code_binary, IdentDB, Unknown
from saml2.mdie import to_dict, from_dict

from saml2 import md
from saml2 import saml
from saml2.extension import mdui
from saml2.extension import idpdisc
from saml2.extension import dri
from saml2.extension import mdattr
from saml2.extension import ui
from saml2 import xmldsig
from saml2 import xmlenc
import six


ONTS = {
    saml.NAMESPACE: saml,
    mdui.NAMESPACE: mdui,
    mdattr.NAMESPACE: mdattr,
    dri.NAMESPACE: dri,
    ui.NAMESPACE: ui,
    idpdisc.NAMESPACE: idpdisc,
    md.NAMESPACE: md,
    xmldsig.NAMESPACE: xmldsig,
    xmlenc.NAMESPACE: xmlenc
}

__author__ = 'rolandh'

logger = logging.getLogger(__name__)


class CorruptDatabase(Exception):
    pass


def context_match(cfilter, cntx):
    # TODO
    return True


class SessionStorageMDB(object):
    """ Session information is stored in a MongoDB database"""

    def __init__(self, database="", collection="assertion", **kwargs):
        db = _mdb_get_database(database, **kwargs)
        self.assertion = db[collection]

    def store_assertion(self, assertion, to_sign):
        name_id = assertion.subject.name_id
        nkey = sha1(code_binary(name_id)).hexdigest()

        doc = {
            "name_id_key": nkey,
            "assertion_id": assertion.id,
            "assertion": to_dict(assertion, ONTS.values(), True),
            "to_sign": to_sign
        }

        _ = self.assertion.insert(doc)

    def get_assertion(self, cid):
        res = []
        for item in self.assertion.find({"assertion_id": cid}):
            res.append({"assertion": from_dict(item["assertion"], ONTS, True),
                        "to_sign": item["to_sign"]})
        if len(res) == 1:
            return res[0]
        elif res is []:
            return None
        else:
            raise SystemError("More then one assertion with the same ID")

    def get_assertions_by_subject(self, name_id=None, session_index=None,
                                  requested_context=None):
        """

        :param name_id: One of name_id or key can be used to get the authn
            statement
        :param session_index: If match against a session index should be done
        :param requested_context: Authn statements should match a specific
            authn context
        :return:
        """
        result = []
        key = sha1(code_binary(name_id)).hexdigest()
        for item in self.assertion.find({"name_id_key": key}):
            assertion = from_dict(item["assertion"], ONTS, True)
            if session_index or requested_context:
                for statement in assertion.authn_statement:
                    if session_index:
                        if statement.session_index == session_index:
                            result.append(assertion)
                            break
                    if requested_context:
                        if context_match(requested_context,
                                         statement.authn_context):
                            result.append(assertion)
                            break
            else:
                result.append(assertion)
        return result

    def remove_authn_statements(self, name_id):
        logger.debug("remove authn about: %s" % name_id)
        key = sha1(code_binary(name_id)).hexdigest()
        for item in self.assertion.find({"name_id_key": key}):
            self.assertion.remove(item["_id"])

    def get_authn_statements(self, name_id, session_index=None,
                             requested_context=None):
        """

        :param name_id:
        :param session_index:
        :param requested_context:
        :return:
        """
        return [k.authn_statement for k in self.get_assertions_by_subject(
            name_id, session_index, requested_context)]


class IdentMDB(IdentDB):
    def __init__(self, database="", collection="ident", domain="",
                 name_qualifier=""):
        IdentDB.__init__(self, None, domain, name_qualifier)
        self.mdb = MDB(database=database, collection=collection)
        self.mdb.primary_key = "user_id"

    def in_store(self, _id):
        if [x for x in self.mdb.get(ident_id=_id)]:
            return True
        else:
            return False

    def create_id(self, nformat, name_qualifier="", sp_name_qualifier=""):
        _id = self._create_id(nformat, name_qualifier, sp_name_qualifier)
        while self.in_store(_id):
            _id = self._create_id(nformat, name_qualifier, sp_name_qualifier)
        return _id

    def store(self, ident, name_id):
        self.mdb.store(ident, name_id=to_dict(name_id, ONTS.values(), True))

    def find_nameid(self, userid, nformat=None, sp_name_qualifier=None,
                    name_qualifier=None, sp_provided_id=None, **kwargs):
        # reset passed for compatibility kwargs for next usage
        kwargs = {}
        if nformat:
            kwargs["name_format"] = nformat
        if sp_name_qualifier:
            kwargs["sp_name_qualifier"] = sp_name_qualifier
        if name_qualifier:
            kwargs["name_qualifier"] = name_qualifier
        if sp_provided_id:
            kwargs["sp_provided_id"] = sp_provided_id

        res = []
        for item in self.mdb.get(userid, **kwargs):
            res.append(from_dict(item["name_id"], ONTS, True))
        return res

    def find_local_id(self, name_id):
        cnid = to_dict(name_id, ONTS.values(), True)
        for item in self.mdb.get(name_id=cnid):
            return item[self.mdb.primary_key]
        return None

    def remove_remote(self, name_id):
        cnid = to_dict(name_id, ONTS.values(), True)
        self.mdb.remove(name_id=cnid)

    def handle_name_id_mapping_request(self, name_id, name_id_policy):
        _id = self.find_local_id(name_id)
        if not _id:
            raise Unknown("Unknown entity")

        if name_id_policy.allow_create == "false":
            raise PolicyError("Not allowed to create new identifier")

        # else create and return a new one
        return self.construct_nameid(_id, name_id_policy=name_id_policy)

    def close(self):
        pass


#------------------------------------------------------------------------------
class MDB(object):
    primary_key = "mdb"

    def __init__(self, database, collection, **kwargs):
        _db = _mdb_get_database(database, **kwargs)
        self.db = _db[collection]

    def store(self, value, **kwargs):
        if value:
            doc = {self.primary_key: value}
        else:
            doc = {}
        doc.update(kwargs)
        _ = self.db.insert(doc)

    def get(self, value=None, **kwargs):
        if value is not None:
            doc = {self.primary_key: value}
            doc.update(kwargs)
            return [item for item in self.db.find(doc)]
        elif kwargs:
            return [item for item in self.db.find(kwargs)]

    def remove(self, key=None, **kwargs):
        if key is None:
            if kwargs:
                for item in self.db.find(kwargs):
                    self.db.remove(item["_id"])
        else:
            doc = {self.primary_key: key}
            doc.update(kwargs)
            for item in self.db.find(doc):
                self.db.remove(item["_id"])

    def keys(self):
        for item in self.db.find():
            yield item[self.primary_key]

    def items(self):
        for item in self.db.find():
            _key = item[self.primary_key]
            del item[self.primary_key]
            del item["_id"]
            yield _key, item

    def __contains__(self, key):
        doc = {self.primary_key: key}
        res = [item for item in self.db.find(doc)]
        if not res:
            return False
        else:
            return True

    def reset(self):
        self.db.drop()


def _mdb_get_database(uri, **kwargs):
    """
    Helper-function to connect to MongoDB and return a database object.

    The `uri' argument should be either a full MongoDB connection URI string,
    or just a database name in which case a connection to the default mongo
    instance at mongodb://localhost:27017 will be made.

    Performs explicit authentication if a username is provided in a connection
    string URI, since PyMongo does not always seem to do that as promised.

    :params database: name as string or (uri, name)
    :returns: pymongo database object
    """
    if not "tz_aware" in kwargs:
        # default, but not forced
        kwargs["tz_aware"] = True

    connection_factory = MongoClient
    _parsed_uri = {}

    try:
        _parsed_uri = pymongo.uri_parser.parse_uri(uri)
    except pymongo.errors.InvalidURI:
        # assume URI to be just the database name
        db_name = uri
        _conn = MongoClient()
        pass
    else:
        if "replicaset" in _parsed_uri["options"]:
            connection_factory = MongoReplicaSetClient
        db_name = _parsed_uri.get("database", "pysaml2")
        _conn = connection_factory(uri, **kwargs)

    _db = _conn[db_name]

    if "username" in _parsed_uri:
        _db.authenticate(
            _parsed_uri.get("username", None),
            _parsed_uri.get("password", None)
        )

    return _db


#------------------------------------------------------------------------------
class EptidMDB(Eptid):
    def __init__(self, secret, database="", collection="eptid"):
        Eptid.__init__(self, secret)
        self.mdb = MDB(database, collection)
        self.mdb.primary_key = "eptid_key"

    def __getitem__(self, key):
        res = self.mdb.get(key)
        if not res:
            raise KeyError(key)
        elif len(res) == 1:
            return res[0]["eptid"]
        else:
            raise CorruptDatabase("Found more than one EPTID document")

    def __setitem__(self, key, value):
        _ = self.mdb.store(key, **{"eptid": value})


#------------------------------------------------------------------------------

def protect(dic):
    res = {}
    for key, val in dic.items():
        key = key.replace(".", "__")
        if isinstance(val, six.string_types):
            pass
        elif isinstance(val, dict):
            val = protect(val)
        elif isinstance(val, list):
            li = []
            for va in val:
                if isinstance(va, six.string_types):
                    pass
                elif isinstance(va, dict):
                    va = protect(va)
                    # I don't think lists of lists will appear am I wrong ?
                li.append(va)
            val = li
        res[key] = val
    return res


def unprotect(dic):
    res = {}
    for key, val in dic.items():
        if key == "__class__":
            pass
        else:
            key = key.replace("__", ".")
        if isinstance(val, six.string_types):
            pass
        elif isinstance(val, dict):
            val = unprotect(val)
        elif isinstance(val, list):
            li = []
            for va in val:
                if isinstance(va, six.string_types):
                    pass
                elif isinstance(val, dict):
                    va = unprotect(va)
                li.append(va)
            val = li
        res[key] = val
    return res


def export_mdstore_to_mongo_db(mds, database, collection, sub_collection=""):
    mdb = MDB(database, collection, sub_collection=sub_collection)
    mdb.reset()
    mdb.primary_key = "entity_id"
    for key, desc in mds.items():
        kwargs = {
            "entity_description": protect(desc),
        }
        mdb.store(key, **kwargs)


class MetadataMDB(InMemoryMetaData):
    def __init__(self, onts, attrc, database="", collection=""):
        super(MetadataMDB, self).__init__(onts, attrc)
        self.mdb = MDB(database, collection)
        self.mdb.primary_key = "entity_id"

    def _ext_service(self, entity_id, typ, service, binding):
        try:
            srvs = self[entity_id][typ]
        except KeyError:
            return None

        if not srvs:
            return srvs

        res = []
        for srv in srvs:
            if "extensions" in srv:
                for elem in srv["extensions"]["extension_elements"]:
                    if elem["__class__"] == service:
                        if elem["binding"] == binding:
                            res.append(elem)

        return res

    def load(self):
        pass

    def items(self):
        for key, item in self.mdb.items():
            yield key, unprotect(item["entity_description"])

    def keys(self):
        return self.mdb.keys()

    def values(self):
        for key, item in self.mdb.items():
            yield unprotect(item["entity_description"])

    def __contains__(self, item):
        return item in self.mdb

    def __getitem__(self, item):
        res = self.mdb.get(item)
        if not res:
            raise KeyError(item)
        elif len(res) == 1:
            return unprotect(res[0]["entity_description"])
        else:
            raise CorruptDatabase("More then one document with key %s" % item)

    def bindings(self, entity_id, typ, service):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import re
from six.moves.urllib.parse import quote_plus
from saml2.httpbase import HTTPBase

from saml2.mdstore import MetadataStore, MetaDataMDX
from saml2.mdstore import destinations
from saml2.mdstore import name

from saml2 import md
from saml2 import sigver
from saml2 import BINDING_SOAP
from saml2 import BINDING_HTTP_REDIRECT
from saml2 import BINDING_HTTP_POST
from saml2 import BINDING_HTTP_ARTIFACT
from saml2 import saml
from saml2 import config
from saml2.attribute_converter import ac_factory
from saml2.attribute_converter import d_to_local_name

from saml2.extension import mdui
from saml2.extension import idpdisc
from saml2.extension import dri
from saml2.extension import mdattr
from saml2.extension import ui
from saml2.s_utils import UnknownPrincipal
from saml2 import xmldsig
from saml2 import xmlenc

from pathutils import full_path

sec_config = config.Config()
#sec_config.xmlsec_binary = sigver.get_xmlsec_binary(["/opt/local/bin"])

ONTS = {
    saml.NAMESPACE: saml,
    mdui.NAMESPACE: mdui,
    mdattr.NAMESPACE: mdattr,
    dri.NAMESPACE: dri,
    ui.NAMESPACE: ui,
    idpdisc.NAMESPACE: idpdisc,
    md.NAMESPACE: md,
    xmldsig.NAMESPACE: xmldsig,
    xmlenc.NAMESPACE: xmlenc
}

ATTRCONV = ac_factory(full_path("attributemaps"))

METADATACONF = {
    "1": [{
        "class": "saml2.mdstore.MetaDataFile",
        "metadata": [(full_path("swamid-1.0.xml"), )],
    }],
    "2": [{
        "class": "saml2.mdstore.MetaDataFile",
        "metadata": [(full_path("InCommon-metadata.xml"), )],
    }],
    "3": [{
        "class": "saml2.mdstore.MetaDataFile",
        "metadata": [(full_path("extended.xml"), )],
    }],
    # "7": [{
    #     "class": "saml2.mdstore.MetaDataFile",
    #     "metadata": [(full_path("metadata_sp_1.xml"), ),
    #                  (full_path("InCommon-metadata.xml"), )], },
    #       {
    #     "class": "saml2.mdstore.MetaDataExtern",
    #     "metadata": [
    #         ("https://kalmar2.org/simplesaml/module.php/aggregator/?id=kalmarcentral2&set=saml2",
    #          full_path("kalmar2.pem")), ],
    # }],
    "4": [{
        "class": "saml2.mdstore.MetaDataFile",
        "metadata": [(full_path("metadata_example.xml"), )],
    }],
    "5": [{
        "class": "saml2.mdstore.MetaDataFile",
        "metadata": [(full_path("metadata.aaitest.xml"), )],
    }],
    "8": [{
        "class": "saml2.mdstore.MetaDataMD",
        "metadata": [(full_path("swamid.md"), )],
    }],
    "9": [{
        "class": "saml2.mdstore.MetaDataFile",
        "metadata": [(full_path("metadata"), )]
    }],
    "10": [{
        "class": "saml2.mdstore.MetaDataExtern",
        "metadata": [
            ("http://md.incommon.org/InCommon/InCommon-metadata-export.xml",
             full_path("inc-md-cert.pem"))]
        }
    ]
}


def _eq(l1, l2):
    return set(l1) == set(l2)


def _fix_valid_until(xmlstring):
    new_date = datetime.datetime.now() + datetime.timedelta(days=1)
    new_date = new_date.strftime("%Y-%m-%dT%H:%M:%SZ")
    return re.sub(r' validUntil=".*?"', ' validUntil="%s"' % new_date,
                  xmlstring)


def test_swami_1():
    UMU_IDP = 'https://idp.umu.se/saml2/idp/metadata.php'
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["1"])
    assert len(mds) == 1  # One source
    idps = mds.with_descriptor("idpsso")
    assert idps.keys()
    idpsso = mds.single_sign_on_service(UMU_IDP)
    assert len(idpsso) == 1
    assert destinations(idpsso) == [
        'https://idp.umu.se/saml2/idp/SSOService.php']

    _name = name(mds[UMU_IDP])
    assert _name == u'Umeå University (SAML2)'
    certs = mds.certs(UMU_IDP, "idpsso", "signing")
    assert len(certs) == 1

    sps = mds.with_descriptor("spsso")
    assert len(sps) == 108

    wants = mds.attribute_requirement('https://connect8.sunet.se/shibboleth')
    lnamn = [d_to_local_name(mds.attrc, attr) for attr in wants["optional"]]
    assert _eq(lnamn, ['eduPersonPrincipalName', 'mail', 'givenName', 'sn',
                       'eduPersonScopedAffiliation'])

    wants = mds.attribute_requirement('https://beta.lobber.se/shibboleth')
    assert wants["required"] == []
    lnamn = [d_to_local_name(mds.attrc, attr) for attr in wants["optional"]]
    assert _eq(lnamn, ['eduPersonPrincipalName', 'mail', 'givenName', 'sn',
                       'eduPersonScopedAffiliation', 'eduPersonEntitlement'])


def test_incommon_1():
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["2"])

    print(mds.entities())
    assert mds.entities() > 1700
    idps = mds.with_descriptor("idpsso")
    print(idps.keys())
    assert len(idps) > 300  # ~ 18%
    try:
        _ = mds.single_sign_on_service('urn:mace:incommon:uiuc.edu')
    except UnknownPrincipal:
        pass

    idpsso = mds.single_sign_on_service('urn:mace:incommon:alaska.edu')
    assert len(idpsso) == 1
    print(idpsso)
    assert destinations(idpsso) == [
        'https://idp.alaska.edu/idp/profile/SAML2/Redirect/SSO']

    sps = mds.with_descriptor("spsso")

    acs_sp = []
    for nam, desc in sps.items():
        if "attribute_consuming_service" in desc:
            acs_sp.append(nam)

    assert len(acs_sp) == 0

    # Look for attribute authorities
    aas = mds.with_descriptor("attribute_authority")

    print(aas.keys())
    assert len(aas) == 180


def test_ext_2():
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["3"])
    # No specific binding defined

    ents = mds.with_descriptor("spsso")
    for binding in [BINDING_SOAP, BINDING_HTTP_POST, BINDING_HTTP_ARTIFACT,
                    BINDING_HTTP_REDIRECT]:
        assert mds.single_logout_service(list(ents.keys())[0], binding, "spsso")


def test_example():
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["4"])
    assert len(mds.keys()) == 1
    idps = mds.with_descriptor("idpsso")

    assert list(idps.keys()) == [
        'http://xenosmilus.umdc.umu.se/simplesaml/saml2/idp/metadata.php']
    certs = mds.certs(
        'http://xenosmilus.umdc.umu.se/simplesaml/saml2/idp/metadata.php',
        "idpsso", "signing")
    assert len(certs) == 1


def test_switch_1():
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["5"])
    assert len(mds.keys()) > 160
    idps = mds.with_descriptor("idpsso")
    print(idps.keys())
    idpsso = mds.single_sign_on_service(
        'https://aai-demo-idp.switch.ch/idp/shibboleth')
    assert len(idpsso) == 1
    print(idpsso)
    assert destinations(idpsso) == [
        'https://aai-demo-idp.switch.ch/idp/profile/SAML2/Redirect/SSO']
    assert len(idps) > 30
    aas = mds.with_descriptor("attribute_authority")
    print(aas.keys())
    aad = aas['https://aai-demo-idp.switch.ch/idp/shibboleth']
    print(aad.keys())
    assert len(aad["attribute_authority_descriptor"]) == 1
    assert len(aad["idpsso_descriptor"]) == 1

    sps = mds.with_descriptor("spsso")
    dual = [eid for eid, ent in idps.items() if eid in sps]
    print(len(dual))
    assert len(dual) == 0


def test_metadata_file():
    sec_config.xmlsec_binary = sigver.get_xmlsec_binary(["/opt/local/bin"])
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["8"])
    print(len(mds.keys()))
    assert len(mds.keys()) == 560


# pyff-test not available
# def test_mdx_service():
#     sec_config.xmlsec_binary = sigver.get_xmlsec_binary(["/opt/local/bin"])
#     http = HTTPBase(verify=False, ca_bundle=None)
#
#     mdx = MetaDataMDX(quote_plus, ONTS.values(), ATTRCONV,
#                       "http://pyff-test.nordu.net",
#                       sec_config, None, http)
#     foo = mdx.service("https://idp.umu.se/saml2/idp/metadata.php",
#                       "idpsso_descriptor", "single_sign_on_service")
#
#     assert len(foo) == 1
#     assert foo.keys()[0] == BINDING_HTTP_REDIRECT
#
#
# def test_mdx_certs():
#     sec_config.xmlsec_binary = sigver.get_xmlsec_binary(["/opt/local/bin"])
#     http = HTTPBase(verify=False, ca_bundle=None)
#
#     mdx = MetaDataMDX(quote_plus, ONTS.values(), ATTRCONV,
#                       "http://pyff-test.nordu.net",
#                       sec_config, None, http)
#     foo = mdx.certs("https://idp.umu.se/saml2/idp/metadata.php", "idpsso")
#
#     assert len(foo) == 1


def test_load_local_dir():
    sec_config.xmlsec_binary = sigver.get_xmlsec_binary(["/opt/local/bin"])
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["9"])
    print(mds)
    assert len(mds) == 3  # Three sources
    assert len(mds.keys()) == 4  # number of idps


def test_load_extern_incommon():
    sec_config.xmlsec_binary = sigver.get_xmlsec_binary(["/opt/local/bin"])
    mds = MetadataStore(ONTS.values(), ATTRCONV, sec_config,
                        disable_ssl_certificate_validation=True)

    mds.imp(METADATACONF["10"])
    print(mds)
    assert mds
    assert len(mds.keys())

if __name__ == "__main__":
    test_load_extern_incommon()



#!/usr/bin/env python

import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


install_requires = [
    # core dependencies
    'decorator',
    'requests >= 1.0.0',
    'paste',
    'zope.interface',
    'repoze.who',
    'pycrypto >= 2.5',  # 'Crypto'
    'pytz',
    'pyOpenSSL',
    'python-dateutil',
    'six'
]

tests_require = [
    'mongodict',
    'pyasn1',
    'pymongo==3.0.1',
    'python-memcached >= 1.51',
    'pytest',
    'mako',
    'webob',
    #'pytest-coverage',
]

setup(
    name='pysaml2',
    version='3.0.0',
    description='Python implementation of SAML Version 2',
    # long_description = read("README"),
    author='Roland Hedberg',
    author_email='roland.hedberg@adm.umu.se',
    license='Apache 2.0',
    url='https://github.com/rohe/pysaml2',

    packages=['saml2', 'saml2/xmldsig', 'saml2/xmlenc', 'saml2/s2repoze',
              'saml2/s2repoze.plugins', "saml2/profile", "saml2/schema",
              "saml2/extension", "saml2/attributemaps", "saml2/authn_context",
              "saml2/entity_category", "saml2/userinfo", "saml2/ws"],

    package_dir={'': 'src'},
    package_data={'': ['xml/*.xml']},
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4"
    ],

    scripts=["tools/parse_xsd2.py", "tools/make_metadata.py",
             "tools/mdexport.py", "tools/merge_metadata.py"],

    tests_require=tests_require,
    extras_require={
        'testing': tests_require,
    },
    install_requires=install_requires,
    zip_safe=False,
    test_suite='tests',
    cmdclass={'test': PyTest},
)



from idp_test import CheckSaml2IntMetaData
from idp_test.check import CheckSaml2IntAttributes
from saml2 import SamlBase, ExtensionContainer

__author__ = 'rolandh'

from idp_test.saml2base import AuthnRequest

class DummyExtension(SamlBase):
    """The urn:mace:umu.se:SAML:2.0:extension:foo element """

    c_tag = 'DummyExtension'
    c_namespace = "urn:mace:umu.se:SAML:2.0:extension:foo"
    c_value_type = {'base': 'NCName'}
    c_children = SamlBase.c_children.copy()
    c_attributes = SamlBase.c_attributes.copy()
    c_child_order = SamlBase.c_child_order[:]
    c_cardinality = SamlBase.c_cardinality.copy()

class AuthnRequest_UnknownIssuer(AuthnRequest):
    def pre_processing(self, message, args):
        _issuer = message.issuer
        _issuer.text = "https://www.example.com/foobar.xml"
        return message

class AuthnRequest_UnknownExtension(AuthnRequest):
    def pre_processing(self, message, args):
        message.extension = ExtensionContainer()
        message.extension.add_extension_element(DummyExtension(text="foo"))
        return message

OPERATIONS = {
    'authn_unknown-issuer': {
        "name": 'AuthnRequest with unknown issuer',
        "descr": 'AuthnRequest with unknown issuer',
        "sequence": [AuthnRequest_UnknownIssuer],
        "depends": ['authn'],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": [CheckSaml2IntAttributes]}
    },
    'authn_unknown-extension': {
        "name": 'AuthnRequest with unknown extension',
        "descr": 'AuthnRequest with unknown extension',
        "sequence": [AuthnRequest_UnknownExtension],
        "depends": ['authn'],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": [CheckSaml2IntAttributes]}
    },
}



__author__ = 'rohe0002'

import json
import logging
import six

from urlparse import urlparse
from bs4 import BeautifulSoup

from mechanize import ParseResponseEx
from mechanize._form import ControlNotFoundError, AmbiguityError
from mechanize._form import ListControl

logger = logging.getLogger(__name__)

def pick_interaction(interactions, _base="", content="", req=None):
    unic = content
    if content:
        _bs = BeautifulSoup(content)
    else:
        _bs = None

    for interaction in interactions:
        _match = 0
        for attr, val in interaction["matches"].items():
            if attr == "url":
                if val == _base:
                    _match += 1
            elif attr == "title":
                if _bs is None:
                    break
                if _bs.title is None:
                    break
                if val in _bs.title.contents:
                    _match += 1
                else:
                    _c = _bs.title.contents
                    if isinstance(_c, list) and not isinstance(
                      _c, six.string_types):
                        for _line in _c:
                            if val in _line:
                                _match += 1
                                continue
            elif attr == "content":
                if unic and val in unic:
                    _match += 1
            elif attr == "class":
                if req and val == req:
                    _match += 1

        if _match == len(interaction["matches"]):
            return interaction

    raise KeyError("No interaction matched")


class FlowException(Exception):
    def __init__(self, function="", content="", url=""):
        Exception.__init__(self)
        self.function = function
        self.content = content
        self.url = url

    def __str__(self):
        return json.dumps(self.__dict__)


class RResponse():
    """
    A Response class that behaves in the way that mechanize expects it.
    Links to a requests.Response
    """
    def __init__(self, resp):
        self._resp = resp
        self.index = 0
        self.text = resp.text
        if isinstance(self.text, unicode):
            if resp.encoding == "UTF-8":
                self.text = self.text.encode("utf-8")
            else:
                self.text = self.text.encode("latin-1")
        self._len = len(self.text)
        self.url = str(resp.url)
        self.statuscode = resp.status_code

    def geturl(self):
        return self._resp.url

    def __getitem__(self, item):
        try:
            return getattr(self._resp, item)
        except AttributeError:
            return getattr(self._resp.headers, item)

    def __getattribute__(self, item):
        try:
            return getattr(self._resp, item)
        except AttributeError:
            return getattr(self._resp.headers, item)

    def read(self, size=0):
        """
        Read from the content of the response. The class remembers what has
        been read so it's possible to read small consecutive parts of the
        content.

        :param size: The number of bytes to read
        :return: Somewhere between zero and 'size' number of bytes depending
            on how much it left in the content buffer to read.
        """
        if size:
            if self._len < size:
                return self.text
            else:
                if self._len == self.index:
                    part = None
                elif self._len - self.index < size:
                    part = self.text[self.index:]
                    self.index = self._len
                else:
                    part = self.text[self.index: self.index + size]
                    self.index += size
                return part
        else:
            return self.text


def pick_form(response, url=None, **kwargs):
    """
    Picks which form in a web-page that should be used

    :param response: A HTTP request response. A DResponse instance
    :param content: The HTTP response content
    :param url: The url the request was sent to
    :return: The picked form or None of no form matched the criteria.
    """
    #_txt = response.text

    forms = ParseResponseEx(response)
    if not forms:
        raise FlowException(content=response.text, url=url)

    #if len(forms) == 1:
    #    return forms[0]
    #else:

    _form = None
    # ignore the first form, because I use ParseResponseEx which adds
    # one form at the top of the list
    forms = forms[1:]
    if len(forms) == 1:
        _form = forms[0]
    else:
        if "pick" in kwargs:
            _dict = kwargs["pick"]
            for form in forms:
                if _form:
                    break
                for key, _ava in _dict.items():
                    if key == "form":
                        _keys = form.attrs.keys()
                        for attr, val in _ava.items():
                            if attr in _keys and val == form.attrs[attr]:
                                _form = form
                    elif key == "control":
                        prop = _ava["id"]
                        _default = _ava["value"]
                        try:
                            orig_val = form[prop]
                            if isinstance(orig_val, six.string_types):
                                if orig_val == _default:
                                    _form = form
                            elif _default in orig_val:
                                _form = form
                        except KeyError:
                            pass
                        except ControlNotFoundError:
                            pass
                    elif key == "method":
                        if form.method == _ava:
                            _form = form
                    else:
                        _form = None

                    if not _form:
                        break
        elif "index" in kwargs:
            _form = forms[int(kwargs["index"])]

    return _form


def do_click(httpc, form, **kwargs):
    """
    Emulates the user clicking submit on a form.

    :param httpc: The Client instance
    :param form: The form that should be submitted
    :return: What do_request() returns
    """

    if "click" in kwargs:
        request = None
        _name = kwargs["click"]
        try:
            _ = form.find_control(name=_name)
            request = form.click(name=_name)
        except AmbiguityError:
            # more than one control with that name
            _val = kwargs["set"][_name]
            _nr = 0
            while True:
                try:
                    cntrl = form.find_control(name=_name, nr=_nr)
                    if cntrl.value == _val:
                        request = form.click(name=_name, nr=_nr)
                        break
                    else:
                        _nr += 1
                except ControlNotFoundError:
                    raise Exception("No submit control with the name='%s' and "
                                    "value='%s' could be found" % (_name,
                                                                   _val))
    else:
        request = form.click()

    headers = {}
    for key, val in request.unredirected_hdrs.items():
        headers[key] = val

    url = request._Request__original

    if form.method == "POST":
        return httpc.send(url, "POST", data=request.data, headers=headers)
    else:
        return httpc.send(url, "GET", headers=headers)


def select_form(httpc, orig_response, **kwargs):
    """
    Pick a form on a web page, possibly enter some information and submit
    the form.

    :param httpc: A HTTP client instance
    :param orig_response: The original response (as returned by requests)
    :return: The response do_click() returns
    """
    response = RResponse(orig_response)
    try:
        _url = response.url
    except KeyError:
        _url = kwargs["location"]

    form = pick_form(response, _url, **kwargs)
    #form.backwards_compatible = False
    if not form:
        raise Exception("Can't pick a form !!")

    if "set" in kwargs:
        for key, val in kwargs["set"].items():
            if key.startswith("_"):
                continue
            if "click" in kwargs and kwargs["click"] == key:
                continue

            try:
                form[key] = val
            except ControlNotFoundError:
                pass
            except TypeError:
                cntrl = form.find_control(key)
                if isinstance(cntrl, ListControl):
                    form[key] = [val]
                else:
                    raise

    return do_click(httpc, form, **kwargs)


#noinspection PyUnusedLocal
def chose(httpc, orig_response, path, **kwargs):
    """
    Sends a HTTP GET to a url given by the present url and the given
    relative path.

    :param orig_response: The original response
    :param content: The content of the response
    :param path: The relative path to add to the base URL
    :return: The response do_click() returns
    """


    if not path.startswith("http"):
        try:
            _url = orig_response.url
        except KeyError:
            _url = kwargs["location"]

        part = urlparse(_url)
        url = "%s://%s%s" % (part[0], part[1], path)
    else:
        url = path

    return httpc.send(url, "GET")
    #return resp, ""


def post_form(httpc, orig_response, **kwargs):
    """
    The same as select_form but with no possibility of change the content
    of the form.

    :param httpc: A HTTP Client instance
    :param orig_response: The original response (as returned by requests)
    :param content: The content of the response
    :return: The response do_click() returns
    """
    response = RResponse(orig_response)

    form = pick_form(response, **kwargs)

    return do_click(httpc, form, **kwargs)


def NoneFunc():
    return None


#noinspection PyUnusedLocal
def parse(httpc, orig_response, **kwargs):
    # content is a form from which I get the SAMLResponse
    response = RResponse(orig_response)

    form = pick_form(response, **kwargs)
    #form.backwards_compatible = False
    if not form:
        raise Exception("Can't pick a form !!")

    return {"SAMLResponse": form["SAMLResponse"],
            "RelayState": form["RelayState"]}


#noinspection PyUnusedLocal
def interaction(args):
    _type = args["type"]
    if _type == "form":
        return select_form
    elif _type == "link":
        return chose
    elif _type == "response":
        return parse
    else:
        return NoneFunc

# ========================================================================


class Operation(object):
    def __init__(self, args=None):
        if args:
            self.function = interaction(args)

        self.args = args or {}
        self.request = None

    def update(self, dic):
        self.args.update(dic)

    #noinspection PyUnusedLocal
    def post_op(self, result, conv, args):
        pass

    def __call__(self, httpc, conv, location, response, content,
                 features):
        try:
            _args = self.args.copy()
        except (KeyError, AttributeError):
            _args = {}

        _args["location"] = location
        _args["features"] = features

        logger.info("--> FUNCTION: %s" % self.function.__name__)
        logger.info("--> ARGS: %s" % _args)

        result = self.function(httpc, response, **_args)
        self.post_op(result, conv, _args)
        return result



from saml2 import samlp
from saml2 import BINDING_HTTP_ARTIFACT
from saml2 import BINDING_HTTP_POST
from saml2 import BINDING_HTTP_REDIRECT
from saml2 import BINDING_PAOS
from saml2 import BINDING_SOAP
from saml2 import BINDING_URI
from saml2.saml import NAMEID_FORMAT_PERSISTENT
from saml2.saml import NAMEID_FORMAT_UNSPECIFIED
from saml2.saml import NAMEID_FORMAT_TRANSIENT
from saml2.saml import NAMEID_FORMAT_EMAILADDRESS

from idp_test.check import CheckLogoutSupport
from idp_test.check import CheckSaml2IntAttributes
from idp_test.check import CheckSaml2IntMetaData
from idp_test.check import VerifyAttributeNameFormat
from idp_test.check import VerifyFunctionality
from idp_test.check import VerifyContent
from idp_test.check import VerifyNameIDMapping
from idp_test.check import VerifyNameIDPolicyUsage
from idp_test.check import VerifySuccessStatus
from idp_test.check import VerifyDigestAlgorithm
from idp_test.check import VerifySignatureAlgorithm
from idp_test.check import VerifySignedPart
from idp_test.check import VerifyEndpoint

from saml2.samlp import NameIDPolicy

__author__ = 'rolandh'


class Request(object):
    _args = {}
    _class = None
    tests = {"post": [VerifyContent], "pre": []}

    def __init__(self, conv):
        self.args = self._args.copy()
        self.conv = conv

    def setup(self):
        pass

    def pre_processing(self, message, args):
        return message

    def post_processing(self, message):
        return message

#class Saml2IntRequest(Request):
#    tests = {"pre": [],
#             "post": [CheckSaml2IntAttributes, VerifyContent
#                      #  CheckSubjectNameIDFormat,
#             ]}


class AuthnRequest(Request):
    _class = samlp.AuthnRequest
    request = "authn_request"
    _args = {"response_binding": BINDING_HTTP_POST,
             "request_binding": BINDING_HTTP_REDIRECT,
             "nameid_format": NAMEID_FORMAT_PERSISTENT,
             "allow_create": True}
    tests = {"pre": [VerifyFunctionality],
             "post": [CheckSaml2IntAttributes,
                      VerifyAttributeNameFormat,
                      VerifySignedPart,
                      VerifyDigestAlgorithm,
                      VerifySignatureAlgorithm]}


class AuthnRequestNID_Transient(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["nameid_format"] = NAMEID_FORMAT_TRANSIENT

    def setup(self):
        cnf = self.conv.client.config
        endps = cnf.getattr("endpoints", "sp")
        url = ""
        for url, binding in endps["assertion_consumer_service"]:
            if binding == BINDING_HTTP_POST:
                self.args["assertion_consumer_service_url"] = url
                break

        self.tests["post"].append((VerifyEndpoint, {"endpoint": url}))


class AuthnRequestNID_Email(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["nameid_format"] = NAMEID_FORMAT_EMAILADDRESS

    def setup(self):
        cnf = self.conv.client.config
        endps = cnf.getattr("endpoints", "sp")
        url = ""
        for url, binding in endps["assertion_consumer_service"]:
            if binding == BINDING_HTTP_POST:
                self.args["assertion_consumer_service_url"] = url
                break

        self.tests["post"].append((VerifyEndpoint, {"endpoint": url}))


class AuthnRequestNID_Unspecified(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["nameid_format"] = NAMEID_FORMAT_UNSPECIFIED

    def setup(self):
        cnf = self.conv.client.config
        endps = cnf.getattr("endpoints", "sp")
        url = ""
        for url, binding in endps["assertion_consumer_service"]:
            if binding == BINDING_HTTP_POST:
                self.args["assertion_consumer_service_url"] = url
                break

        self.tests["post"].append((VerifyEndpoint, {"endpoint": url}))


class AuthnRequestNID_no(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["nameid_format"] = ""

    def setup(self):
        cnf = self.conv.client.config
        endps = cnf.getattr("endpoints", "sp")
        url = ""
        for url, binding in endps["assertion_consumer_service"]:
            if binding == BINDING_HTTP_POST:
                self.args["assertion_consumer_service_url"] = url
                break

        self.tests["post"].append((VerifyEndpoint, {"endpoint": url}))


class AuthnRequestEndpointIndex(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["attribute_consuming_service_index"] = 3

    def setup(self):
        cnf = self.conv.client.config
        endps = cnf.getattr("endpoints", "sp")
        acs3 = endps["assertion_consumer_service"][3]
        self.tests["post"].append((VerifyEndpoint, {"endpoint": acs3[0]}))


class AuthnRequestEndpointIndexNIDTransient(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["attribute_consuming_service_index"] = 3
        self.args["nameid_format"] = NAMEID_FORMAT_TRANSIENT

    def setup(self):
        cnf = self.conv.client.config
        endps = cnf.getattr("endpoints", "sp")
        acs3 = endps["assertion_consumer_service"][3]
        self.tests["post"].append((VerifyEndpoint, {"endpoint": acs3[0]}))


class AuthnRequestSpecEndpoint(AuthnRequest):
    def setup(self):
        cnf = self.conv.client.config
        endps = cnf.getattr("endpoints", "sp")
        acs3 = endps["assertion_consumer_service"][3]
        self.args["assertion_consumer_service_url"] = acs3[0]
        self.tests["post"].append((VerifyEndpoint, {"endpoint": acs3[0]}))


class DynAuthnRequest(Request):
    _class = samlp.AuthnRequest
    request = "authn_request"
    _args = {"response_binding": BINDING_HTTP_POST}
    tests = {}
    name_id_formats = [NAMEID_FORMAT_TRANSIENT, NAMEID_FORMAT_PERSISTENT]
    bindings = [BINDING_HTTP_REDIRECT, BINDING_HTTP_POST]

    def setup(self):
        metadata = self.conv.client.metadata
        entity = metadata[self.conv.entity_id]
        self.args.update({"nameid_format": "", "request_binding": ""})
        for idp in entity["idpsso_descriptor"]:
            for nformat in self.name_id_formats:
                if self.args["nameid_format"]:
                    break
                for nif in idp["name_id_format"]:
                    if nif["text"] == nformat:
                        self.args["nameid_format"] = nformat
                        break
            for bind in self.bindings:
                if self.args["request_binding"]:
                    break
                for sso in idp["single_sign_on_service"]:
                    if sso["binding"] == bind:
                        self.args["request_binding"] = bind
                        break


class AuthnRequestPost(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["request_binding"] = BINDING_HTTP_POST


class AuthnRequest_using_Artifact(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["response_binding"] = BINDING_HTTP_ARTIFACT
        self.args["binding"] = BINDING_HTTP_ARTIFACT


class AuthnRequest_using_ArtifactNID_Transient(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["nameid_format"] = NAMEID_FORMAT_TRANSIENT
        self.args["response_binding"] = BINDING_HTTP_ARTIFACT
        self.args["binding"] = BINDING_HTTP_ARTIFACT


class AuthnRequestPostNID_Transient(AuthnRequestPost):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["nameid_format"] = NAMEID_FORMAT_TRANSIENT


class LogOutRequest(Request):
    request = "logout_request"
    _args = {"request_binding": BINDING_SOAP}
    tests = {"pre": [VerifyFunctionality], "post": []}

    def __init__(self, conv):
        Request.__init__(self, conv)
        self.tests["pre"].append(CheckLogoutSupport)
        #self.tests["post"].append(VerifyLogout)

    def setup(self):
        resp = self.conv.saml_response[-1].response
        assertion = resp.assertion[0]
        subj = assertion.subject
        self.args["name_id"] = subj.name_id
        self.args["issuer_entity_id"] = assertion.issuer.text


class AssertionIDRequest(Request):
    request = "assertion_id_request"
    _args = {"request_binding": BINDING_URI,
             "response_binding": None}
    tests = {"pre": [VerifyFunctionality]}

    def setup(self):
        assertion = self.conv.saml_response[-1].assertion
        self.args["assertion_id_refs"] = [assertion.id]


class AuthnQuery(Request):
    request = "authn_query"
    _args = {"request_binding": BINDING_SOAP}
    tests = {"pre": [VerifyFunctionality], "post": []}

    def __init__(self, conv):
        Request.__init__(self, conv)
        self.tests["post"].append(VerifySuccessStatus)

    def setup(self):
        assertion = self.conv.saml_response[-1].assertion
        self.args["subject"] = assertion.subject


class NameIDMappingRequest(Request):
    request = "name_id_mapping_request"
    _args = {"request_binding": BINDING_SOAP,
             "name_id_policy": NameIDPolicy(format=NAMEID_FORMAT_PERSISTENT,
                                            sp_name_qualifier="GroupOn",
                                            allow_create="true")}

    def __init__(self, conv):
        Request.__init__(self, conv)
        self.tests["post"].append(VerifyNameIDMapping)

    def setup(self):
        assertion = self.conv.saml_response[-1].assertion
        self.args["name_id"] = assertion.subject.name_id


class AuthnRequest_NameIDPolicy1(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["name_id_policy"] = NameIDPolicy(
            format=NAMEID_FORMAT_PERSISTENT, sp_name_qualifier="Group1",
            allow_create="true")
        self.tests["post"].append(VerifyNameIDPolicyUsage)


class AuthnRequest_NameIDPolicy1Transient(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["name_id_policy"] = NameIDPolicy(
            format=NAMEID_FORMAT_TRANSIENT, sp_name_qualifier="Group1",
            allow_create="true")
        self.args["nameid_format"] = NAMEID_FORMAT_TRANSIENT
        self.tests["post"].append(VerifyNameIDPolicyUsage)


class AuthnRequest_TransientNameID(AuthnRequest):
    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["name_id_policy"] = NameIDPolicy(
            format=NAMEID_FORMAT_TRANSIENT, sp_name_qualifier="Group",
            allow_create="true")
        self.tests["post"].append(VerifyNameIDPolicyUsage)


class ECP_AuthnRequest(AuthnRequest):

    def __init__(self, conv):
        AuthnRequest.__init__(self, conv)
        self.args["request_binding"] = BINDING_SOAP
        self.args["service_url_binding"] = BINDING_PAOS

    def setup(self):
        _client = self.conv.client
        _client.user = "babs"
        _client.passwd = "howes"

    # def post_processing(self, message):
    #     # Unpacking SOAP message
    #     return parse_soap_enveloped_saml_response(message)


class ManageNameIDRequest(Request):
    request = "manage_name_id_request"
    _args = {"request_binding": BINDING_SOAP,
             "new_id": samlp.NewID("New identifier")}

    def __init__(self, conv):
        Request.__init__(self, conv)
        self.tests["post"].append(VerifySuccessStatus)

    def setup(self):
        assertion = self.conv.saml_response[-1].assertion
        self.args["name_id"] = assertion.subject.name_id


class AttributeQuery(Request):
    request = "attribute_query"
    _args = {"request_binding": BINDING_SOAP}
    tests = {"pre": [VerifyFunctionality],
             "post": [CheckSaml2IntAttributes, VerifyAttributeNameFormat]}

    def setup(self):
        assertion = self.conv.saml_response[-1].assertion
        self.args["name_id"] = assertion.subject.name_id

# -----------------------------------------------------------------------------

OPERATIONS = {
    'verify': {
        'tc_id': "S2c-16",
        "name": 'Verify SAML connectivity',
        "descr": 'Uses AuthnRequest to check connectivity',
        "sequence": [DynAuthnRequest],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []}
    },
    'authn': {
        "tc_id": "S2c-02",
        "name": 'Absolute basic AuthnRequest',
        "descr": 'AuthnRequest using HTTP-Redirect',
        "sequence": [AuthnRequest],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []},
        "depend":["verify"]
    },
    'authn-nid_transient': {
        "tc_id": "S2c-10",
        "name": 'AuthnRequest, NameID-trans',
        "descr": 'Basic SAML2 AuthnRequest, HTTP-Redirect, '
                 'transient name ID',
        "sequence": [AuthnRequestNID_Transient],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []},
        "depend":["authn"]
    },
    'authn-nid_email': {
        "tc_id": "S2c-20",
        "name": 'AuthnRequest email nameID',
        "descr": 'Basic SAML2 AuthnRequest, HTTP-Redirect, NameID-email'
                 'specified',
        "sequence": [AuthnRequestNID_Email],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []},
        "depend":["authn"]
    },
    'authn-nid_no': {
        "tc_id": "S2c-21",
        "name": 'AuthnRequest no NameID format',
        "descr": 'Basic SAML2 AuthnRequest, HTTP-Redirect, no NameID format '
                 'specified',
        "sequence": [AuthnRequestNID_no],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []},
        "depend":["authn"]
    },
    'authn-nid_unspecified': {
        "tc_id": "S2c-21",
        "name": 'AuthnRequest using unspecified NameID format',
        "descr": 'Basic SAML2 AuthnRequest, HTTP-Redirect, NameID-unspec',
        "sequence": [AuthnRequestNID_Unspecified],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []},
        "depend":["authn"]
    },
    'authn-post': {
        "tc_id": "S2c-08",
        "name": 'Basic SAML2 AuthnRequest using HTTP POST',
        "descr": 'AuthnRequest using HTTP-POST',
        "sequence": [AuthnRequestPost],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []},
        "depend":["authn"]
    },
    'authn-post-transient': {
        "tc_id": "S2c-09",
        "name": 'AuthnRequest HTTP-POST, transient NameID fmt',
        "descr": 'AuthnRequest using HTTP-POST expecting transient NameID',
        "sequence": [AuthnRequestPostNID_Transient],
        "tests": {"pre": [CheckSaml2IntMetaData],
                  "post": []},
        "depend":["authn-post"]
    },
    'attribute-query':{
        "tc_id": "S2c-01",
        "name": "Attribute query",
        "sequence":[AuthnRequest, AttributeQuery],
        "depend":["authn"]
    },
    'attribute-query-transient':{
        "tc_id": "S2c-20",
        "name": "Attribute query, NameID transient",
        "sequence":[AuthnRequestNID_Transient, AttributeQuery],
        "depend":["authn"]
    },
    'authn_endpoint_index': {
        "tc_id": "S2c-03",
        "name": 'AuthnRequest, endpoint index',
        "descr": '',
        "sequence": [AuthnRequestEndpointIndex],
        "depend":["authn"]
    },
    'authn_endpoint_index-transient': {
        "tc_id": "S2c-03",
        "name": 'AuthnRequest, endpoint index, NameID-trans',
        "descr": '',
        "sequence": [AuthnRequestEndpointIndexNIDTransient],
        "depend":["authn"]
    },
    'authn_specified_endpoint': {
        "tc_id": "S2c-04",
        "name": 'AuthnRequest, specified endpoint',
        "descr": '',
        "sequence": [AuthnRequestSpecEndpoint],
        "depend":["authn"]
    },
    'authn-artifact':{
       'tc_id': "S2c-05",
       "name": "SAML2 AuthnRequest using an artifact",
       "descr": ('AuthnRequest using HTTP-Redirect and artifact'),
       "sequence": [AuthnRequest_using_Artifact]
    },
    'authn-artifact_nid-transient':{
       'tc_id': "S2c-05",
       "name": "SAML2 AuthnRequest expecting artifact response",
       "descr": ('AuthnRequest using HTTP-Redirect and artifact'),
       "sequence": [AuthnRequest_using_ArtifactNID_Transient]
    },
    'authn-assertion_id_request': {
        "tc_id": "S2c-06",
        "name": 'AuthnRequest then AssertionIDRequest',
        "descr": 'AuthnRequest followed by an AssertionIDRequest',
        "sequence": [AuthnRequest, AssertionIDRequest],
        "tests": {"pre": [CheckSaml2IntMetaData],  "post": []},
        "depend":["authn"]
    },
    'authn-nid_transient-assertion_id_request': {
        "tc_id": "S2c-26",
        "name": 'AuthnRequest then AssertionIDRequest, NameID-trans',
        "descr": 'AuthnRequest followed by an AssertionIDRequest',
        "sequence": [AuthnRequestNID_Transient, AssertionIDRequest],
        "tests": {"pre": [CheckSaml2IntMetaData],  "post": []},
        "depend":["authn"]
    },
    'authn-with-name_id_policy': {
        "tc_id": "S2c-11",
        "name": 'SAML2 AuthnRequest with specific NameIDPolicy',
        "descr": 'AuthnRequest with specific NameIDPolicy',
        "sequence": [AuthnRequest_NameIDPolicy1],
        "tests": {"pre": [CheckSaml2IntMetaData],  "post": []},
        "depend":["authn"]
    },
    'authn-with-name_id_policy_nid-transient': {
        "tc_id": "S2c-31",
        "name": 'AuthnRequest NameIDPolicy transient',
        "descr": 'AuthnRequest with specific NameIDPolicy',
        "sequence": [AuthnRequest_NameIDPolicy1Transient],
        "tests": {"pre": [CheckSaml2IntMetaData],  "post": []},
        "depend":["authn"]
    },
    'ecp_authn': {
        'tc_id': "S2c-12",
        "name": "AuthnRequest using ECP and PAOS",
        "descr": "SAML2 AuthnRequest using ECP and PAOS",
        "sequence":[ECP_AuthnRequest]
    },
    'log-in-out': {
        "tc_id": "S2c-13",
        "name": 'Basic SAML2 log in and out',
        "descr": 'AuthnRequest using HTTP-Redirect followed by a logout',
        "sequence": [AuthnRequest, LogOutRequest],
        "tests": {"pre": [CheckSaml2IntMetaData],  "post": []},
        "depend":["authn"]
    },
    'manage_nameid':{
        "tc_id": "S2c-14",
        "name":  "ManageNameID; set NameID",
        "descr": "Setting the SP provided ID by using ManageNameID",
        "sequence":[AuthnRequest, ManageNameIDRequest],
        "depend":["authn"]
    },
    'nameid-mapping':{
        "tc_id": "S2c-15",
        "name": "Simple NameIDMapping request",
        "sequence":[AuthnRequest, NameIDMappingRequest],
        "depend":["authn"]
    },
    'manage_nameid_nid-transient':{
        "tc_id": "S2c-16",
        "name": "ManageNameID; set NameID; AuthRequ/NameID-trans",
        "descr": "Setting the SP provided ID by using ManageNameID",
        "sequence":[AuthnRequestNID_Transient, ManageNameIDRequest],
        "depend":["authn"]
    },
    'authn-authn_query': {
        "tc_id": "S2c-17",
        "name": 'AuthnRequest then AuthnQuery',
        "descr": 'AuthnRequest followed by an AuthnQuery',
        "sequence": [AuthnRequest, AuthnQuery],
        "tests": {"pre": [CheckSaml2IntMetaData],  "post": []},
        "depend":["authn"]
    },
}


from __future__ import print_function
import json
import pprint
import argparse
import os.path
import sys
import traceback
from importlib import import_module

from idp_test import SCHEMA
from saml2 import root_logger

from saml2.mdstore import MetadataStore, MetaData
from saml2.saml import NAME_FORMAT_UNSPECIFIED
from saml2.server import Server
from saml2.config import IdPConfig
from saml2.config import logging

from sp_test.base import Conversation

from saml2test import FatalError
from saml2test import CheckError
from saml2test import ContextFilter
from saml2test import exception_trace
from saml2test.check import CRITICAL

__author__ = 'rolandh'

#formatter = logging.Formatter("%(asctime)s %(name)s:%(levelname)s %(message)s")
formatter_2 = logging.Formatter(
    "%(delta).6f - %(levelname)s - [%(name)s] %(message)s")

cf = ContextFilter()
cf.start()

streamhandler = logging.StreamHandler(sys.stderr)
streamhandler.setFormatter(formatter_2)

memoryhandler = logging.handlers.MemoryHandler(1024 * 10, logging.DEBUG)
memoryhandler.addFilter(cf)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(memoryhandler)
logger.setLevel(logging.DEBUG)


class Client(object):
    def __init__(self, operations, check_factory):
        self.operations = operations
        self.tests = None
        self.check_factory = check_factory

        self._parser = argparse.ArgumentParser()
        self._parser.add_argument("-c", dest="config", default="config",
                                  help="Configuration file for the IdP")
        self._parser.add_argument(
            '-C', dest="ca_certs",
            help=("CA certs to use to verify HTTPS server certificates, ",
                  "if HTTPS is used and no server CA certs are defined then ",
                  "no cert verification will be done"))
        self._parser.add_argument('-d', dest='debug', action='store_true',
                                  help="Print debug information")
        self._parser.add_argument("-H", dest="pretty", action='store_true',
                                  help="Output summary on stdout as pretty "
                                       "printed python dict instead of JSON")
        self._parser.add_argument("-i", dest="insecure", action='store_true',
                                  help="Do not verify SSL certificate")
        self._parser.add_argument("-I", dest="keysdir", default="keys",
                                  help="Directory for invalid IDP keys")
        self._parser.add_argument('-J', dest="json_config_file",
                                  help="Test target configuration")
        self._parser.add_argument(
            '-k', dest='content_log', action='store_true',
            help="Log HTTP content in spearate files in directory "
                 "<operation>/, which defaults to the path in -L")
        self._parser.add_argument(
            "-l", dest="list", action="store_true",
            help="List all the test flows as a JSON object")
        self._parser.add_argument("-L", dest="logpath", default=".",
                                  help="Path to the logfile directory")
        self._parser.add_argument('-m', dest="metadata", action='store_true',
                                  help="Return the IdP metadata")
        self._parser.add_argument(
            "-P", dest="configpath", default=".",
            help="Path to the configuration file for the IdP")
        self._parser.add_argument("-t", dest="testpackage",
                                  help="Module describing tests")
        #self._parser.add_argument('-v', dest='verbose', action='store_true',
        #                          help="Print runtime information") # unsused
        self._parser.add_argument("-Y", dest="pysamllog", action='store_true',
                                  help="Print PySAML2 logs")
        self._parser.add_argument("oper", nargs="?", help="Which test to run")

        self.interactions = None
        self.entity_id = None
        self.constraints = {}
        self.args = None
        self.idp = None
        self.idp_config = None

    def json_config_file(self):
        if self.args.json_config_file == "-":
            return json.loads(sys.stdin.read())
        else:
            return json.loads(open(self.args.json_config_file).read())

    def idp_configure(self, metadata_construction=False):
        sys.path.insert(0, self.args.configpath)
        mod = import_module(self.args.config)
        self.idp_config = IdPConfig().load(mod.CONFIG, metadata_construction)

        if not self.args.insecure:
            self.idp_config.verify_ssl_cert = False
        else:
            if self.args.ca_certs:
                self.idp_config.ca_certs = self.args.ca_certs
            else:
                self.idp_config.ca_certs = "../keys/cacert.pem"
        # hack to change idp cert without config change. TODO: find interface to
        # change IDP cert after __init__
        if self.args.oper == 'sp-04':
            self.idp_config.cert_file = os.path.join(self.args.keysdir, "non_md_cert.pem")
            self.idp_config.key_file = os.path.join(self.args.keysdir, "non_md_key.pem")
            for f in [self.idp_config.cert_file, self.idp_config.key_file]:
                if not os.path.isfile(f):
                    print("File not found: %s" % os.path.abspath(f))
                    raise

        self.idp = Server(config=self.idp_config)

    def test_summation(self, sid):
        status = 0
        for item in self.test_log:
            if item["status"] > status:
                status = item["status"]

        if status == 0:
            status = 1

        info = {
            "id": sid,
            "status": status,
            "tests": self.test_log
        }

        if status == 5:
            info["url"] = self.test_log[-1]["url"]
            info["htmlbody"] = self.test_log[-1]["message"]

        return info

    def output_log(self, memhndlr, hndlr2):
        """
        """

        print(80 * ":", file=sys.stderr)
        hndlr2.setFormatter(formatter_2)
        memhndlr.setTarget(hndlr2)
        memhndlr.flush()
        memhndlr.close()

    def run(self):
        self.args = self._parser.parse_args()

        if self.args.pysamllog:
            root_logger.addHandler(memoryhandler)
            root_logger.setLevel(logging.DEBUG)

        if self.args.metadata:
            return self.make_meta()
        elif self.args.list:
            return self.list_operations()
        elif self.args.oper == "check":
            return self.verify_metadata()
        else:
            if not self.args.oper:
                raise Exception("Missing test case specification")
            self.args.oper = self.args.oper.strip("'")
            self.args.oper = self.args.oper.strip('"')

        self.setup()

        try:
            oper = self.operations.OPERATIONS[self.args.oper]
        except KeyError:
            if self.tests:
                try:
                    oper = self.tests.OPERATIONS[self.args.oper]
                except ValueError:
                    print("Undefined testcase " + self.args.oper,
                          file=sys.stderr)
                    return
            else:
                print("Undefined testcase " + self.args.oper, file=sys.stderr)
                return

        if self.args.pretty:
            pp = pprint.PrettyPrinter(indent=4)
        else:
            pp = None

        logger.info("Starting conversation")
        conv = Conversation(self.idp, self.idp_config,
                            self.interactions, self.json_config,
                            check_factory=self.check_factory,
                            entity_id=self.entity_id,
                            constraints=self.constraints,
                            commandlineargs = self.args)
        try:
            conv.do_sequence_and_tests(oper["sequence"], oper["tests"])
            self.test_log = conv.test_output
            tsum = self.test_summation(self.args.oper)
            err = None
        except CheckError as err:
            self.test_log = conv.test_output
            tsum = self.test_summation(self.args.oper)
        except FatalError as err:
            if conv:
                self.test_log = conv.test_output
                self.test_log.append(exception_trace("RUN", err))
            else:
                self.test_log = exception_trace("RUN", err)
            tsum = self.test_summation(self.args.oper)
        except Exception as err:
            if conv:
                conv.test_output.append({"status": CRITICAL,
                         "name": "test driver error",
                         "id": "critial exception"})
                self.test_log = conv.test_output
                self.test_log.append(exception_trace("RUN", err))
            else:
                self.test_log = exception_trace("RUN", err)
            tsum = self.test_summation(self.args.oper)
            logger.error("Unexpected exception in test driver %s" %
                         traceback.format_exception(*sys.exc_info()))


        if pp:
            pp.pprint(tsum)
        else:
            print(json.dumps(tsum), file=sys.stdout)

        if tsum["status"] > 1 or self.args.debug or err:
            self.output_log(memoryhandler, streamhandler)

    def setup(self):
        self.json_config = self.json_config_file()

        _jc = self.json_config

        try:
            self.interactions = _jc["interaction"]
        except KeyError:
            self.interactions = []

        self.idp_configure()

        metadata = MetadataStore(SCHEMA, self.idp_config.attribute_converters,
                                 self.idp_config)
        info = _jc["metadata"].encode("utf-8")
        md = MetaData(SCHEMA, self.idp_config.attribute_converters, info)
        md.load()
        metadata[0] = md
        self.idp.metadata = metadata
        #self.idp_config.metadata = metadata

        if self.args.testpackage:
            self.tests = import_module("sp_test.package.%s" %
                                       self.args.testpackage)

        try:
            self.entity_id = _jc["entity_id"]
            # Verify its the correct metadata
            assert self.entity_id in md.entity.keys()
        except KeyError:
            if len(md.entity.keys()) == 1:
                self.entity_id = md.entity.keys()[0]
            else:
                raise Exception("Don't know which entity to talk to")

        if "constraints" in _jc:
            self.constraints = _jc["constraints"]
            if "name_format" not in self.constraints:
                self.constraints["name_format"] = NAME_FORMAT_UNSPECIFIED

    def make_meta(self):
        pass

    def list_operations(self):
        res = []
        for key, val in self.operations.OPERATIONS.items():
            res.append({"id": key, "name": val["name"]})

        print(json.dumps(res))

    def verify_metadata(self):
        pass



#!/usr/bin/env python

import argparse

from saml2 import saml
from saml2 import md
from saml2 import xmldsig
from saml2 import xmlenc

from saml2.attribute_converter import ac_factory
from saml2.httpbase import HTTPBase
from saml2.extension import dri
from saml2.extension import idpdisc
from saml2.extension import mdattr
from saml2.extension import mdrpi
from saml2.extension import mdui
from saml2.extension import shibmd
from saml2.extension import ui

from saml2.sigver import _get_xmlsec_cryptobackend
from saml2.sigver import SecurityContext

from saml2.mdstore import MetaDataFile
from saml2.mdstore import MetaDataExtern

__author__ = 'rolandh'

"""
A script that imports and verifies metadata.
"""


ONTS = {
    saml.NAMESPACE: saml,
    mdui.NAMESPACE: mdui,
    mdattr.NAMESPACE: mdattr,
    mdrpi.NAMESPACE: mdrpi,
    dri.NAMESPACE: dri,
    ui.NAMESPACE: ui,
    idpdisc.NAMESPACE: idpdisc,
    md.NAMESPACE: md,
    xmldsig.NAMESPACE: xmldsig,
    xmlenc.NAMESPACE: xmlenc,
    shibmd.NAMESPACE: shibmd
}


parser = argparse.ArgumentParser()
parser.add_argument('-t', dest='type')
parser.add_argument('-u', dest='url')
parser.add_argument('-c', dest='cert')
parser.add_argument('-a', dest='attrsmap')
parser.add_argument('-o', dest='output')
parser.add_argument('-x', dest='xmlsec')
parser.add_argument('-i', dest='ignore_valid', action='store_true')
parser.add_argument(dest="item")
args = parser.parse_args()


metad = None

if args.ignore_valid:
    kwargs = {"check_validity": False}
else:
    kwargs = {}

if args.type == "local":
    if args.cert and args.xmlsec:
        crypto = _get_xmlsec_cryptobackend(args.xmlsec)
        sc = SecurityContext(crypto)
        metad = MetaDataFile(ONTS.values(), args.item, args.item,
                             cert=args.cert, security=sc, **kwargs)
    else:
        metad = MetaDataFile(ONTS.values(), args.item, args.item, **kwargs)
elif args.type == "external":
    ATTRCONV = ac_factory(args.attrsmap)
    httpc = HTTPBase()
    crypto = _get_xmlsec_cryptobackend(args.xmlsec)
    sc = SecurityContext(crypto)
    metad = MetaDataExtern(ONTS.values(), ATTRCONV, args.url,
                           sc, cert=args.cert, http=httpc, **kwargs)

if metad:
    try:
        metad.load()
    except:
        raise
    else:
        print("OK")



